/*********************************************************************

時計型の妖怪音声おもちゃ(白)用 妖怪コードの赤外線リモコン制御ソフト

				IR to Yok Code Converter for Arduino

　　　お手持ちのシャープのテレビ用の赤外線リモコンの数字ボタンで
　　　４ケタの妖怪コードを入力すると動作します。
　　　また、チャンネル選局の上ボタンで、次のコードを呼び出します。

本ソースリストおよびソフトウェアは、ライセンスフリーです。
利用、編集、再配布等が自由に行えますが、著作権表示の改変は禁止します。

                               Copyright (c) 2015 Wataru KUNINO
                               http://www.geocities.jp/bokunimowakaru/
*********************************************************************/
/*
・赤外線リモコン受信モジュールを Arduino Digital 2 に接続します。
・妖怪おもちゃ内部基板のSW3～6を Arduino Digital 3～6 に接続します。
・赤外線リモコンを使用しない場合は#defineのIrRemoENを0にします。

注意点：
　赤外線リモコンを使用する(IrRemoENが1)状態の時は低消費電力モードで動作します。
　この場合、シリアルからのYokCodeを受信することが出来ません。
　シリアルから妖怪ウォッチを制御する場合は、IrRemoENを0に設定してください。

参考文献：
　妖怪コードに関する情報は下記のサイトを参考にしました。
　http://yokai.kaathemachine.com/
*/

#define SW3_PIN			3						// 妖怪おもちゃ内SW3の接続先
#define SW4_PIN			4						// 妖怪おもちゃ内SW4の接続先
#define SW5_PIN			5						// 妖怪おもちゃ内SW5の接続先
#define SW6_PIN			6						// 妖怪おもちゃ内SW6の接続先
#define LED_PIN			13						// 動作確認用LEDのDigitalピン番号
#define IR_VCC_PIN		8						// 赤外線モジュール用電源ピン
//#define IR_PIN		2 (固定)				// 赤外線センサ(Digital 2 固定)

#define IrRemoEN		1						// 赤外線リモコン入力 1:使用
#define SerialEN		1						// シリアル入力 1:使用
#define SymWait			10						// YokCode出力間隔
#define IR_DATA_LEN		8						// リモコンコードのデータ長(byte)
#define AutoPowerSave	8						// 自動で電源を切るまでの時間(8秒倍)

#include <avr/sleep.h>
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

volatile int keyValue = 0;						// ユーザ入力用バッファ
volatile byte keyLength = 0;					// ユーザ入力文字数
volatile byte ardIsrSrc = 1;					// 割込み状態
volatile boolean yokCodeMedal = 0;				// YokCodeメダルの状態 0:空
volatile unsigned int time = 0;					// 時間カウント用
int yokCode = 2913;								// 初期のYokCode(2913=地縛霊猫)

void port_off(){
// 全SWポートをオープンにする
	pinMode(SW3_PIN,INPUT);
	pinMode(SW4_PIN,INPUT);
	pinMode(SW5_PIN,INPUT);
	pinMode(SW6_PIN,INPUT);
}

void port_on(){
// 全SWポートをLOW出力につする
	pinMode(SW3_PIN,OUTPUT);
	pinMode(SW4_PIN,OUTPUT);
	pinMode(SW5_PIN,OUTPUT);
	pinMode(SW6_PIN,OUTPUT);
	digitalWrite(SW3_PIN,LOW);
	digitalWrite(SW4_PIN,LOW);
	digitalWrite(SW5_PIN,LOW);
	digitalWrite(SW6_PIN,LOW);
}

void port_out(byte in){
// ポートにYokCodeを出力する
	digitalWrite(SW3_PIN,in>>0 & 0x01);
	digitalWrite(SW4_PIN,in>>1 & 0x01);
	digitalWrite(SW5_PIN,in>>2 & 0x01);
	digitalWrite(SW6_PIN,in>>3 & 0x01);
}

void code_out(int code,boolean dir){
	byte out,i;
	
	port_on();
	if(SerialEN){
		Serial.print("yokCode = "); Serial.print(code,DEC);
		Serial.print(", dir = ");  Serial.println(dir,DEC);
	}
	if(!dir){
		port_out(0x0A); delay(SymWait);
		port_out(0x00); delay(SymWait);
	}
	for(i=0;i<3;i++){
		if(dir){	// dir=1:メダル挿入
			out = (byte)(code>>(4*(2-i))) & 0x0F;
		}else{		// dir=0:メダル取り出し
			out = (byte)(code>>(4*i)) & 0x0F;
		}
		port_out(out);
		delay(SymWait);
		port_out(0x00); delay(SymWait);
	}
	if(dir){
		port_out(0x0A); delay(SymWait);
		port_out(0x00); delay(SymWait);
	}
	port_off();
}


int code_increment(int code){
	byte data[3],flag=1,i;
	
	if(code >= 3511) return(2382);
	for(i=0;i<3;i++) data[i]=(byte)(code>>(4*(int)i)) & 0x0F;
	while(flag){
		data[0]++;
		if(data[0] > 0x0F){
			data[0] = 0x00;
			data[1]++;
			if(data[1] > 0x0F){
				data[1] = 0x00;
				data[2]+= 0x02;
			}
		}
		if(	data[0]!=data[2] &&
			data[0]!=data[1] &&
			data[0]!=0x00 &&
			data[0]!=0x05 &&
			data[0]!=0x0A
		) flag=0;
	}
	code=0;
	for(i=0;i<3;i++){
		code *= 16;
		code += (int)data[2-i];
	}
	if(SerialEN){
		Serial.print("IncCode = 0x");
		for(i=0;i<3;i++) Serial.print(data[2-i],HEX);
		Serial.print(" -> "); Serial.println(code,DEC);
	}
	return(code);
}

int code_decrement(int code){
	byte data[3],flag=1,i;
	
	if(code <= 2382) return(3511);
	for(i=0;i<3;i++) data[i]=(byte)(code>>(4*(int)i)) & 0x0F;
	while(flag){
		data[0]--;
		if(data[0] > 0x0F){
			data[0] = 0x0F;
			data[1]--;
			if(data[1] > 0x0F){
				data[1] = 0x0F;
				data[2]-= 0x02;
			}
		}
		if(	data[0]!=data[2] &&
			data[0]!=data[1] &&
			data[0]!=0x00 &&
			data[0]!=0x05 &&
			data[0]!=0x0A
		) flag=0;
	}
	code=0;
	for(i=0;i<3;i++){
		code *= 16;
		code += (int)data[2-i];
	}
	if(SerialEN){
		Serial.print("DecCode = 0x");
		for(i=0;i<3;i++) Serial.print(data[2-i],HEX);
		Serial.print(" -> "); Serial.println(code,DEC);
	}
	return(code);
}

byte ir2dec4sharp(byte *data){	// data長＝6バイト
	byte ret=0xFF;
	if(data[0]==0xAA && data[1]==0x5A && data[2]==0x8F && data[3]==0x12){
		ret=(byte)'\n';
		if(data[4] >= 0x01 && data[4] <= 0x09) ret = data[4];
		if(data[4] >= 0x4E && data[4] <= 0x56) ret = data[4] - 0x4D;
		if(data[4] == 0x0A || data[4] == 0x57) ret = 0x00;
		if(data[4] == 0x11) ret=(byte)'+';
		if(data[4] == 0x12) ret=(byte)'-';
	}
	if(SerialEN){
		Serial.print("RemCode = ");
		for(byte i=0;i<6;i++){
			if(data[i]<0x10)Serial.print('0');
			Serial.print(data[i],HEX); Serial.print(' ');
		}
		Serial.print("-> ");
		if(ret<10) Serial.println(ret,DEC); else Serial.println("\\n");
	}
	return(ret);
}

void error(){
	while(1){digitalWrite(LED_PIN,1);delay(200);digitalWrite(LED_PIN,0);delay(200);}
}

void setup(){
	pinMode(LED_PIN,OUTPUT); 
	pinMode(IR_VCC_PIN,OUTPUT);
	digitalWrite(IR_VCC_PIN,HIGH);
	port_off();
	if(!(IrRemoEN|SerialEN)) error();
	if(IrRemoEN) ir_read_init();
	if(SerialEN){
		Serial.begin(9600);
		Serial.println("IR to Yok Code Converter by Wataru KUNINO");
		delay(100);
	}
}

void loop(){
	byte data[IR_DATA_LEN];
	byte keyIn=0xFF;
	
	digitalWrite(LED_PIN,HIGH);
	if(IrRemoEN && !ardIsrSrc){
		if(ir_read(data,IR_DATA_LEN,0) > 8){		// 8ビット以上の赤外線データを読み込む
			keyIn = ir2dec4sharp(data);
		}
		do delay(100);								// 受信信号が無くなるまで待つ
		while(ir_read(data,1,0)>0);					// (チャタリング防止のため)
		ardIsrSrc=1;
		time = 0;
	}
	if(SerialEN){
		if(Serial.available()) keyIn = (byte)Serial.read() - (byte)'0';
	}
	if(IrRemoEN && ardIsrSrc==2){					// WDT起動
		if(SerialEN){
			Serial.print("WDTimer = ");	Serial.print(time,DEC);
			Serial.print("/");	Serial.println(AutoPowerSave,DEC);
			delay(20);
		}
		if(time > AutoPowerSave){					// オートパワーオフ
			digitalWrite(IR_VCC_PIN,LOW);
			digitalWrite(LED_PIN,LOW);
			while(1) system_sleep();
		}
	}
	
	if(keyIn != 0xFF){
		if(keyIn < 10){								// 数字入力時
			keyValue *= 10;
			keyValue += (int)keyIn;
			keyLength++;
		}else if(keyIn == (byte)'\n'){				// 実行コマンドの入力時
			keyLength = 0; keyValue = 0;
			yokCodeMedal = 0;						// メダル空状態に設定
			code_out(yokCode,0);					// メダル取り外し実行
		}else if(keyIn == (byte)'+'){
			if(yokCodeMedal){
				code_out(yokCode,0); delay(1500);
			}
			yokCode=code_increment(yokCode);		// 次のコードを取得
			code_out(yokCode,1);					// メダル挿入の実行
			yokCodeMedal = 1;						// メダル挿入状態に設定
		}else if(keyIn == (byte)'-'){	
			if(yokCodeMedal){
				code_out(yokCode,0); delay(1500);
			}
			yokCode=code_decrement(yokCode);		// 前のコードを取得
			code_out(yokCode,1);					// メダル挿入の実行
			yokCodeMedal = 1;						// メダル挿入状態に設定
		}
	}
	if(keyLength >= 4){								// 4桁入力 かつ メダル非挿入
		if(yokCodeMedal){							// メダル挿入状態の時
			code_out(yokCode,0); delay(1500);		// メダル取り外し実行
		}
		yokCode = keyValue;
		code_out(yokCode,1);						// メダル挿入の実行
		yokCodeMedal = 1;							// メダル挿入状態に設定
		keyLength = 0; keyValue = 0;
	}
	
	/* Arduino電源オフ */
	digitalWrite(LED_PIN,LOW);
	if(IrRemoEN){
		setup_watchdog(9);							// WDTを割り込みモード(8秒)で起動
		attachInterrupt(0, ardIsrIr, CHANGE);		// Digital 2を割込み入力に設定
		interrupts();
		system_sleep();								// Arduino パワーダウン
	}
}

void ardIsrIr(){
	noInterrupts();
	ardIsrSrc=0;
}

ISR(WDT_vect) {									// WDTがタイムアップした時に実行される処理
	noInterrupts();
	time++;
	ardIsrSrc=2;
}
